/* arg formatting by line, taken from https://github.com/sphinx-doc/sphinx/issues/1514#issuecomment-742703082 */


/*Newlines (\a) and spaces (\20) before each parameter*/
dl.class em:not([class])::before {
    content: "\a\20\20\20\20\20\20\20\20\20\20\20\20\20\20\20\20";
    white-space: pre;
}

/*Newline after the last parameter (so the closing bracket is on a new line)*/
dl.class em:not([class]):last-of-type::after {
    content: "\a";
    white-space: pre;
}

/*To have blue background of width of the block (instead of width of content)*/
dl.class > dt:first-of-type {
    display: block !important;
}

button.copybtn {
    height:25px;
    width:25px;
    opacity: 0.5;
    padding: 0;
    border: none;
    background: none;
}
